try:
    import matplotlib
    matplotlib.use('Agg')
    from matplotlib.backends.backend_pdf import PdfPages
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

from PyPDF2 import PdfFileMerger, PdfFileReader

import os

import click
from scipy.stats import gaussian_kde
from numpy import linspace, concatenate, around

def color_blind_friendly(color_palette):
    
    color_dict = {"normal":["#F5793A", "#0F2080"], "protan":["#AE9C45", "#052955"], "deutran":["#C59434", "#092C48"], "tritan":["#1ECBE1", "#E1341E"]}
    
    if color_palette not in color_dict:
        click.echo(f"WARN: {color_palette} is not a valid color_palette, must be one of 'normal'. 'protan', 'deutran', or 'tritan'. Using default 'normal'.")
        color_palette = "normal"
    return color_dict[color_palette][0], color_dict[color_palette][1]

def save_report(pdf_path, title, top_decoys, top_targets, cutoffs, svalues, qvalues, pvalues, pi0, color_palette="normal"):

    if plt is None:
        raise ImportError("Error: The matplotlib package is required to create a report.")

    t_col, d_col = color_blind_friendly(color_palette)

    plt.figure(figsize=(10, 15))
    plt.subplots_adjust(hspace=.5)

    plt.subplot(321)
    plt.title("q-/s-value")
    plt.xlabel('false positive rate (q-value)')
    plt.ylabel('true positive rate (s-value)')

    plt.scatter(qvalues, svalues, s=3)
    plt.plot(qvalues, svalues)

    plt.subplot(322)
    plt.title('d-score performance')
    plt.xlabel('d-score cutoff')
    plt.ylabel('rates')

    plt.scatter(cutoffs, svalues, color=t_col, s=3)
    plt.plot(cutoffs, svalues, color=t_col, label="TPR (s-value)")
    plt.scatter(cutoffs, qvalues, color=d_col, s=3)
    plt.plot(cutoffs, qvalues, color=d_col, label="FPR (q-value)")

    plt.subplot(323)
    plt.title("group d-score distributions")
    plt.xlabel("d-score")
    plt.ylabel("# of groups")
    plt.hist(
        [top_targets, top_decoys], 20, color=[t_col, d_col], label=['target', 'decoy'], histtype='bar')
    plt.legend(loc=2)

    plt.subplot(324)
    tdensity = gaussian_kde(top_targets)
    tdensity.covariance_factor = lambda: .25
    tdensity._compute_covariance()
    ddensity = gaussian_kde(top_decoys)
    ddensity.covariance_factor = lambda: .25
    ddensity._compute_covariance()
    xs = linspace(min(concatenate((top_targets, top_decoys))), max(
        concatenate((top_targets, top_decoys))), 200)
    plt.title("group d-score densities")
    plt.xlabel("d-score")
    plt.ylabel("density")
    plt.plot(xs, tdensity(xs), color=t_col, label='target')
    plt.plot(xs, ddensity(xs), color=d_col, label='decoy')
    plt.legend(loc=2)

    plt.subplot(325)
    if pvalues is not None:
        counts, __, __ = plt.hist(pvalues, bins=20, density=True)
        plt.plot([0, 1], [pi0['pi0'], pi0['pi0']], "r")
        plt.title("p-value density histogram: $\pi_0$ = " + str(around(pi0['pi0'], decimals=3)))
        plt.xlabel("p-value")
        plt.ylabel("density histogram")

    if pi0['pi0_smooth'] is not False:
        plt.subplot(326)
        plt.plot(pi0['lambda_'], pi0['pi0_lambda'], ".")
        plt.plot(pi0['lambda_'], pi0['pi0_smooth'], "r")
        plt.xlim([0,1])
        plt.ylim([0,1])
        plt.title("$\pi_0$ smoothing fit plot")
        plt.xlabel("$\lambda$")
        plt.ylabel("$\pi_0$($\lambda$)")

    plt.suptitle(title)
    plt.savefig(pdf_path)

def plot_scores(df, out, color_palette="normal"):

    if plt is None:
        raise ImportError("Error: The matplotlib package is required to create a report.")

    score_columns = ["SCORE"] + [c for c in df.columns if c.startswith("MAIN_VAR_")] + [c for c in df.columns if c.startswith("VAR_")]

    t_col, d_col = color_blind_friendly(color_palette)

    with PdfPages(out) as pdf:
        for idx in score_columns:
            top_targets = df[df["DECOY"] == 0][idx]
            top_decoys = df[df["DECOY"] == 1][idx]

            if not (top_targets.isnull().values.any() or top_targets.isnull().values.any()):
                plt.figure(figsize=(10, 10))
                plt.subplots_adjust(hspace=.5)

                plt.subplot(211)
                plt.title(idx)
                plt.xlabel(idx)
                plt.ylabel("# of groups")
                plt.hist(
                    [top_targets, top_decoys], 20, color=[t_col, d_col], label=['target', 'decoy'], histtype='bar')
                plt.legend(loc=2)

                try:
                    tdensity = gaussian_kde(top_targets)
                    tdensity.covariance_factor = lambda: .25
                    tdensity._compute_covariance()
                    ddensity = gaussian_kde(top_decoys)
                    ddensity.covariance_factor = lambda: .25
                    ddensity._compute_covariance()
                    xs = linspace(min(concatenate((top_targets, top_decoys))), max(
                        concatenate((top_targets, top_decoys))), 200)
                    plt.subplot(212)
                    plt.xlabel(idx)
                    plt.ylabel("density")
                    plt.plot(xs, tdensity(xs), color=t_col, label='target')
                    plt.plot(xs, ddensity(xs), color=d_col, label='decoy')
                    plt.legend(loc=2)
                except:
                    plt.subplot(212)

                pdf.savefig()
                plt.close()

def plot_hist(x, title, xlabel, ylabel, pdf_path="histogram_plot.png"):

    if plt is not None:
        # Clear figures
        plt.close('all')
        counts, __, __ = plt.hist(x, bins=20, density=True)
        plt.title(title, wrap=True)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.savefig(pdf_path)

        # Clear figures
        plt.close('all')

def main_score_selection_report(title, sel_column, mapper, decoy_scores, target_scores, target_pvalues, pi0, color_palette="normal", pdf_path="main_score_selection_report.pdf", worker_num=1):
    
    if plt is None:
        raise ImportError("Error: The matplotlib package is required to create a report.")
    
    # Create output to merge pdges
    output = PdfFileMerger() 
    # Generate colors
    t_col, d_col = color_blind_friendly(color_palette)

    # Clear figures
    plt.close('all')

    plt.figure(figsize=(10, 11))
    plt.subplots_adjust(hspace=.5)

    # Plot Score Distribution
    plt.subplot(121)
    plt.hist([target_scores, decoy_scores], 20, color=[t_col, d_col], label=['target', 'decoy'], histtype='bar')
    plt.title(f"histogram of scores")
    plt.xlabel("score")
    plt.ylabel("density histogram")
    plt.legend(loc=1)
    # Plot P-value Distribution
    plt.subplot(122)
    if target_pvalues is not None:
        counts, __, __ = plt.hist(target_pvalues, bins=20, density=True)
        if pi0 is not None:
            plt.plot([0, 1], [pi0['pi0'], pi0['pi0']], "r")
            plt.title("p-value density histogram: $\pi_0$ = " + str(around(pi0['pi0'], decimals=3)))
        else:
            plt.title("p-value density histogram: $\pi_0$ estimation failed.")
        plt.xlabel("target p-values")
        plt.ylabel("density histogram")
    # Finalize figure
    plt.suptitle(f"{title}: {mapper[sel_column]}")
    plt.tight_layout()
    # Append to existing master report if exists, otherwise write to a new master report pdf.
    if os.path.isfile(pdf_path):
        temp_pdf_path = f"temp_main_score_selection_report_thread__{worker_num}.pdf"
        # Save current plot in temporary pdf
        plt.savefig(temp_pdf_path)
        # Append master pdf and temp pdf to output merger
        output.append(PdfFileReader(open(pdf_path, "rb")))
        output.append(PdfFileReader(open(temp_pdf_path, "rb")))
        # Write to master pdf
        output.write(pdf_path)
        # Remove temporary pdf
        os.remove(temp_pdf_path)
    else:
        # Save as master pdf
        plt.savefig(pdf_path)
    # Clear figures
    plt.close('all')


import sys
import numpy
from setuptools import setup, find_packages
from distutils.extension import Extension

try:
    from Cython.Build import cythonize
except ImportError:
    use_cython = False
else:
    use_cython = True

cmdclass = {}
ext_modules = []

if use_cython:
    ext_modules += [Extension("pyprophet._optimized", ["pyprophet/_optimized.pyx"])]
    ext_modules = cythonize(ext_modules)
else:
    ext_modules += [Extension("pyprophet._optimized", ["pyprophet/_optimized.c"])]

# read the contents of README for PyPI
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='pyprophet',
      version="2.2.1",
      author="The PyProphet Developers",
      author_email="rocksportrocker@gmail.com",
      description="PyProphet: Semi-supervised learning and scoring of OpenSWATH results.",
      long_description=long_description,
      long_description_content_type='text/markdown',
      license="BSD",
      url="https://github.com/PyProphet/pyprophet",
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      include_dirs=[numpy.get_include()],
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Topic :: Scientific/Engineering :: Bio-Informatics',
          'Topic :: Scientific/Engineering :: Chemistry',
      ],
      zip_safe=False,
      install_requires=[
          "Click",
          "numpy >= 1.9.0",
          "scipy",
          "pandas >= 0.17",
          "cython",
          "numexpr >= 2.1",
          "scikit-learn >= 0.17",
          "xgboost",
          "hyperopt",
          "statsmodels >= 0.8.0",
          "matplotlib",
          "tabulate",
          "PyPDF2"
      ],
      entry_points={
          'console_scripts': [
              "pyprophet=pyprophet.main:cli",
              ]
      },
      ext_modules=ext_modules,
      )



cimport cython
cimport libc.stdlib
cimport numpy as np

import numpy as np
import operator


ctypedef np.float32_t DATA_TYPE


@cython.boundscheck(False)
@cython.wraparound(False)
def find_nearest_matches(DATA_TYPE[:] basis, DATA_TYPE[:] sample_points, use_sort_order=1):
    cdef size_t num_basis = basis.shape[0]
    cdef size_t num_samples = sample_points.shape[0]
    result = np.zeros((num_samples,), dtype=np.int64)
    cdef np.int64_t[:] view = result
    cdef size_t i, best_j
    cdef size_t low, high, mid
    cdef DATA_TYPE sp_i, best_dist, dist
    cdef int sort_order

    if not use_sort_order:
        for i in range(num_samples):
            sp_i = sample_points[i]
            best_j = 0
            best_dist = abs(basis[0] - sp_i)
            for j in range(1, num_basis):
                dist = abs(basis[j] - sp_i)
                if dist < best_dist:
                    best_dist = dist
                    best_j = j
            view[i] = best_j
        return result

    sort_order = find_sort_order(basis)
    for i in range(num_samples):
        sp_i = sample_points[i]
        if sort_order == 0:
            best_j = 0
            best_dist = abs(basis[0] - sp_i)
            for j in range(1, num_basis):
                dist = abs(basis[j] - sp_i)
                if dist < best_dist:
                    best_dist = dist
                    best_j = j
        elif sort_order == 1:
            low = 0
            high = num_basis - 1
            best_j = -1
            if basis[low] == sp_i:
                best_j = low
            elif basis[high] == sp_i:
                best_j = high
            else:
                while low < high - 1:
                    mid = (low + high) / 2
                    if basis[mid] == sp_i:
                        best_j = mid
                    if basis[mid] < sp_i:
                        low = mid
                    else:
                        high = mid
                if best_j == -1:
                    if abs(basis[low] - sp_i) < abs(basis[high] - sp_i):
                        best_j = low
                    else:
                        best_j = high
            # find first match in list !
            while best_j > 0:
                if basis[best_j - 1] == basis[best_j]:
                    best_j = best_j - 1
                else:
                    break
        else:
            low = 0
            high = num_basis - 1
            best_j = -1
            if basis[low] == sp_i:
                best_j = low
            elif basis[high] == sp_i:
                best_j = high
            else:
                while low < high - 1:
                    mid = (low + high) / 2
                    if basis[mid] == sp_i:
                        best_j = mid
                        break
                    if basis[mid] > sp_i:
                        low = mid
                    else:
                        high = mid
                if best_j == -1:
                    if abs(basis[low] - sp_i) < abs(basis[high] - sp_i):
                        best_j = low
                    else:
                        best_j = high
            # find first match in list:
            while best_j > 0:
                if basis[best_j - 1] == basis[best_j]:
                    best_j = best_j - 1
                else:
                    break

        view[i] = best_j
    return result


cdef int find_sort_order(DATA_TYPE[:] basis):
    # 0: unsorted
    # 1: ascending
    # 2: descending
    cdef size_t i, n
    n = basis.shape[0]
    if n <= 1:
        return 0
    i = 0
    while i < n - 1 and basis[i] == basis[i + 1]:
        i += 1
    if i == n - 1:
        return 1   # or -1 as list is constant
    if basis[i] < basis[i + 1]:
        for i in range(i, n - 1):
            if basis[i] > basis[i + 1]:
                return 0
        return 1
    else:
        for i in range(i, n - 1):
            if basis[i] < basis[i + 1]:
                return 0
        return -1


@cython.boundscheck(False)
@cython.wraparound(False)
def count_num_positives(np.float64_t[:] values):
    cdef size_t n = values.shape[0]
    cdef np.float64_t[:] inp_view = values
    cdef size_t i0, i1, c
    result = np.zeros_like(values, dtype=np.int64)
    cdef np.int64_t[:] res_view = result
    i0 = i1 = 0
    c = n
    while i0 < n:
        while i1 < n and inp_view[i0] == inp_view[i1]:
            res_view[i1] = c
            i1 += 1
        c -= 1
        i0 += 1
    return result


@cython.boundscheck(False)
@cython.wraparound(False)
def find_top_ranked(np.int64_t[:] tg_ids, DATA_TYPE[:] scores):
    cdef size_t n = scores.shape[0]
    flags = np.zeros((n,), dtype=np.int64)
    cdef np.int64_t[:] view = flags
    cdef DATA_TYPE current_max = scores[0]
    cdef size_t current_imax = 0
    cdef size_t current_tg_id = tg_ids[0]
    cdef size_t current_write_i = 0
    cdef size_t i
    cdef size_t id_
    cdef DATA_TYPE sc
    for i in range(tg_ids.shape[0]):
        id_ = tg_ids[i]
        sc = scores[i]
        if id_ != current_tg_id:
            current_tg_id = id_
            view[current_imax] = 1
            current_write_i += 1
            current_max = sc
            current_imax = i
            continue
        if sc > current_max:
            current_max = sc
            current_imax = i
    view[current_imax] = 1
    return flags


cdef partial_rank(DATA_TYPE[:] scores, size_t imin, size_t imax, np.uint32_t[:] ranks, np.uint32_t * ix):
    """ imax is exclusive, imax-imin is the size of a target group, so 32 bit int should
    be sufficient"""
    cdef np.uint32_t i, j, pos
    for i in range(imax - imin):
        ix[i] = i
    for i in range(imax - imin - 1):
        pos = i
        for j in range(i + 1, imax - imin):
            if scores[ix[j] + imin] > scores[ix[pos] + imin]:
                pos = j
        ix[i], ix[pos] = ix[pos], ix[i]

    for j in range(imax - imin):
        ranks[ix[j] + imin] = j + 1


@cython.boundscheck(False)
@cython.wraparound(False)
def rank(np.int64_t[:] tg_ids, DATA_TYPE[:] scores):
    # we assume a transition group has always max 1000 entries:
    cdef np.uint32_t * ix = <np.uint32_t * > libc.stdlib.malloc(1000 * sizeof(np.uint32_t))
    cdef size_t n = tg_ids.shape[0]
    result = np.zeros((n,), dtype=np.uint32)
    cdef np.uint32_t[:] ranks = result
    cdef size_t imin = 0
    cdef size_t imax
    cdef np.int64_t g0
    while imin < n:
        imax = imin + 1
        g0 = tg_ids[imin]
        while imax < n and tg_ids[imax] == g0:
            imax += 1
        partial_rank(scores, imin, imax, ranks, ix)
        imin = imax
    libc.stdlib.free(ix)
    return result


@cython.boundscheck(False)
@cython.wraparound(False)
def rank32(np.uint32_t[:] tg_ids, DATA_TYPE[:] scores):
    # we assume a transition group has always max 1000 entries:
    cdef np.uint32_t * ix = <np.uint32_t * > libc.stdlib.malloc(1000 * sizeof(np.uint32_t))
    cdef size_t n = tg_ids.shape[0]
    result = np.zeros((n,), dtype=np.uint32)
    cdef np.uint32_t[:] ranks = result
    cdef size_t imin = 0
    cdef size_t imax
    cdef np.float32_t g0
    while imin < n:
        imax = imin + 1
        g0 = tg_ids[imin]
        while imax < n and tg_ids[imax] == g0:
            imax += 1
        partial_rank(scores, imin, imax, ranks, ix)
        imin = imax
    libc.stdlib.free(ix)
    return result


@cython.boundscheck(False)
@cython.wraparound(False)
def single_chromatogram_hypothesis_fast(np.float64_t[:] inv_pg_pp_true, np.float64_t prior_chrom_null, np.float64_t prior_pg_true):

    # inv_pg_pp_true contains the probabilities that the peaks are false

    cdef size_t n = inv_pg_pp_true.shape[0]
    cdef np.float64_t final_val
    cdef np.float64_t val
    cdef np.float64_t sum_all = 0.0
    result = np.zeros((n+1,), dtype=np.float64)
    cdef np.float64_t[:] cresult = result

    cdef int i
    cdef int j
    cdef double af = 1.0

    # Compute the null hypothesis (e.g. all peaks are false)
    i = 0
    while i < n:
        af *= inv_pg_pp_true[i]
        i += 1

    cresult[0] = af * prior_chrom_null

    # Compute the alternative hypotheses (e.g. one peak is true and all other
    # peaks are false)
    i = 0
    while i < n:
        val = inv_pg_pp_true[i]
        j = 0
        af = 1.0
        # Compute the product of all peaks being false except i
        while j < n:
            if i != j:
                af *= inv_pg_pp_true[j]
            j += 1

        final_val = <np.float64_t>( (1-val) * af) # value i is true

        cresult[i+1] = final_val * prior_pg_true
        i += 1

    # Compute sum over all hypothesis
    i = 0
    while i < n+1:
        sum_all += cresult[i]
        i += 1

    # Divide by sum over all hypothesis (normalize)
    i = 0
    while i < n+1:
        cresult[i] = cresult[i] / sum_all
        i += 1

    return result
